#!/usr/bin/env python3
"""
Sui (粋) WebAssembly Runtime
Execute Sui code directly via WebAssembly using wasmtime
"""

import sys
import os

try:
    from wasmtime import Store, Module, Func, FuncType, ValType, Linker, Engine
    WASMTIME_AVAILABLE = True
except ImportError:
    WASMTIME_AVAILABLE = False

from sui2wasm import Sui2WatTranspiler

def get_version() -> str:
    """Get package version"""
    try:
        from importlib.metadata import version
        return version("sui-lang")
    except Exception:
        return "0.4.1"


class SuiWasmRuntime:
    """Runtime for executing Sui code via WebAssembly"""

    def __init__(self):
        if not WASMTIME_AVAILABLE:
            raise RuntimeError("wasmtime is required. Install with: pip install wasmtime")
        
        self.engine = Engine()
        self.output: list[int] = []

    def run(self, sui_code: str) -> tuple[int, list[int]]:
        """
        Execute Sui code via WebAssembly
        
        Args:
            sui_code: Sui source code
            
        Returns:
            Tuple of (return_value, output_list)
        """
        self.output = []
        
        # Transpile Sui to WAT
        transpiler = Sui2WatTranspiler()
        wat_code = transpiler.transpile(sui_code)
        
        # Create store and compile module
        store = Store(self.engine)
        
        try:
            module = Module(self.engine, wat_code)
        except Exception as e:
            raise RuntimeError(f"WAT compilation error: {e}\n\nGenerated WAT:\n{wat_code}")
        
        # Create linker and define imports
        linker = Linker(self.engine)
        
        # Define print_i32 function
        print_type = FuncType([ValType.i32()], [])
        def print_i32(value: int):
            self.output.append(value)
            print(value)
        linker.define_func("env", "print_i32", print_type, print_i32)
        
        # Instantiate module
        try:
            instance = linker.instantiate(store, module)
        except Exception as e:
            raise RuntimeError(f"Wasm instantiation error: {e}")
        
        # Get and call main function
        main_func = instance.exports(store).get("main")
        if main_func is None:
            raise RuntimeError("No main function exported")
        
        result = main_func(store)
        
        return result, self.output

    def run_file(self, filename: str) -> tuple[int, list[int]]:
        """Execute a Sui file via WebAssembly"""
        with open(filename, 'r') as f:
            code = f.read()
        return self.run(code)


def main():
    if len(sys.argv) >= 2 and sys.argv[1] in ('--version', '-V'):
        print(f"sui-lang {get_version()}")
        return
    
    if not WASMTIME_AVAILABLE:
        print("Error: wasmtime is required")
        print("Install with: pip install wasmtime")
        sys.exit(1)

    if len(sys.argv) < 2 or sys.argv[1] in ('--help', '-h'):
        print("Sui (粋) WebAssembly Runtime")
        print("=" * 50)
        print("")
        print("Usage:")
        print("  suiwasm <file.sui>       # Execute Sui file via WebAssembly")
        print("  suiwasm --wat <file.sui> # Show generated WAT code")
        print("  suiwasm --version        # Show version")
        print("")
        print("This runtime:")
        print("  1. Transpiles Sui code to WebAssembly Text Format (WAT)")
        print("  2. Compiles WAT to WebAssembly binary")
        print("  3. Executes in wasmtime runtime")
        print("")
        print("Sample:")
        print("-" * 50)

        sample = """
= v0 1
: 0
> v1 v0 5
? v1 1
. v0
+ v0 v0 1
@ 0
: 1
"""
        print("Sui (print 1-5):")
        print(sample.strip())
        print("")
        print("Executing via WebAssembly:")
        print("-" * 50)
        
        try:
            runtime = SuiWasmRuntime()
            result, output = runtime.run(sample)
            print("-" * 50)
            print(f"Return value: {result}")
        except Exception as e:
            print(f"Error: {e}")
        
        return

    # Check for --wat flag
    if sys.argv[1] == '--wat':
        if len(sys.argv) < 3:
            print("Error: Please specify a file")
            sys.exit(1)
        
        filename = sys.argv[2]
        with open(filename, 'r') as f:
            code = f.read()
        
        transpiler = Sui2WatTranspiler()
        wat_code = transpiler.transpile(code)
        print(wat_code)
        return

    # Execute file
    filename = sys.argv[1]
    
    if not os.path.exists(filename):
        print(f"Error: File not found: {filename}")
        sys.exit(1)

    try:
        runtime = SuiWasmRuntime()
        result, output = runtime.run_file(filename)
        # Return value is shown only if non-zero or if there's no output
        if result != 0 or not output:
            print(f"[Return: {result}]")
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


if __name__ == '__main__':
    main()




"""Tests for Sui to WAT transpiler (sui2wasm.Sui2WatTranspiler)"""

import pytest
import sys
import os

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from sui2wasm import Sui2WatTranspiler


class TestSui2WatBasic:
    """Test basic WAT transpilation"""

    def test_assignment(self):
        """Test variable assignment"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= g0 42")
        assert "(i32.const 42)" in code
        assert "(global.set $g0)" in code

    def test_local_assignment(self):
        """Test local variable assignment"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= v0 10")
        assert "(i32.const 10)" in code
        assert "(local.set $v0)" in code

    def test_arithmetic_add(self):
        """Test addition"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= v0 5\n= v1 3\n+ v2 v0 v1")
        assert "(i32.add)" in code
        assert "(local.set $v2)" in code

    def test_arithmetic_sub(self):
        """Test subtraction"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("- v0 10 3")
        assert "(i32.sub)" in code

    def test_arithmetic_mul(self):
        """Test multiplication"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("* v0 4 5")
        assert "(i32.mul)" in code

    def test_arithmetic_div(self):
        """Test division"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("/ v0 20 4")
        assert "(i32.div_s)" in code

    def test_arithmetic_mod(self):
        """Test modulo"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("% v0 17 5")
        assert "(i32.rem_s)" in code


class TestSui2WatComparisons:
    """Test comparison operations"""

    def test_less_than(self):
        """Test less than comparison"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("< v0 3 5")
        assert "(i32.lt_s)" in code

    def test_greater_than(self):
        """Test greater than comparison"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("> v0 5 3")
        assert "(i32.gt_s)" in code

    def test_equality(self):
        """Test equality comparison"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("~ v0 5 5")
        assert "(i32.eq)" in code


class TestSui2WatLogic:
    """Test logical operations"""

    def test_not(self):
        """Test NOT operation"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= v0 0\n! v1 v0")
        assert "(i32.eqz)" in code

    def test_and(self):
        """Test AND operation"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("& v0 1 1")
        assert "(i32.and)" in code

    def test_or(self):
        """Test OR operation"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("| v0 0 1")
        assert "(i32.or)" in code


class TestSui2WatOutput:
    """Test output operation"""

    def test_print(self):
        """Test print operation imports external function"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(". 42")
        assert '(import "env" "print_i32"' in code
        assert "(call $print_i32)" in code


class TestSui2WatLabelsJumps:
    """Test labels and jumps (state machine pattern)"""

    def test_label_creates_state(self):
        """Test that labels create states"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(": 0\n: 1")
        assert "(local $_state i32)" in code
        assert "(block $exit" in code
        assert "(loop $loop" in code

    def test_unconditional_jump(self):
        """Test unconditional jump"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(": 0\n@ 0")
        assert "(local.set $_state" in code
        assert "(br $loop)" in code

    def test_conditional_jump(self):
        """Test conditional jump"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= v0 1\n: 0\n? v0 1\n: 1")
        assert "(if" in code
        assert "(br $loop)" in code

    def test_loop_pattern(self):
        """Test a typical loop pattern"""
        sui_code = """
= v0 0
: 0
< v1 v0 10
! v2 v1
? v2 1
+ v0 v0 1
@ 0
: 1
"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(sui_code)
        # Should have state machine structure
        assert "State 0" in code or "_state" in code
        assert "(loop $loop" in code


class TestSui2WatArrays:
    """Test array operations"""

    def test_array_creates_memory(self):
        """Test that array operations create memory"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("[ g0 5")
        assert "(memory 1)" in code
        assert "(global $heap_ptr" in code

    def test_array_create(self):
        """Test array creation"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("[ g0 10")
        assert "(global.get $heap_ptr)" in code
        assert "(global.set $heap_ptr)" in code

    def test_array_write(self):
        """Test array write"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("[ g0 5\n{ g0 0 42")
        assert "(i32.store)" in code

    def test_array_read(self):
        """Test array read"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("[ g0 5\n{ g0 0 42\n] v0 g0 0")
        assert "(i32.load)" in code


class TestSui2WatFunctions:
    """Test function definitions and calls"""

    def test_function_definition(self):
        """Test function definition"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("# 0 1 {\n^ a0\n}")
        assert "(func $f0" in code
        assert "(param $a0 i32)" in code
        assert "(result i32)" in code

    def test_function_with_multiple_params(self):
        """Test function with multiple parameters"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("# 0 3 {\n+ v0 a0 a1\n+ v1 v0 a2\n^ v1\n}")
        assert "(param $a0 i32)" in code
        assert "(param $a1 i32)" in code
        assert "(param $a2 i32)" in code

    def test_function_call(self):
        """Test function call"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("# 0 1 {\n^ a0\n}\n$ g0 0 42")
        assert "(call $f0)" in code

    def test_return_statement(self):
        """Test return statement"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("# 0 1 {\n^ a0\n}")
        assert "(return)" in code


class TestSui2WatModuleStructure:
    """Test overall module structure"""

    def test_module_wrapper(self):
        """Test that output is wrapped in module"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= g0 1")
        assert code.startswith("(module")
        assert code.strip().endswith(")")

    def test_main_function_export(self):
        """Test that main function is exported"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= g0 1")
        assert '(func $main (export "main")' in code

    def test_main_returns_i32(self):
        """Test that main returns i32"""
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile("= g0 1")
        assert "(result i32)" in code


class TestSui2WatExamples:
    """Test that example files transpile without errors"""

    @pytest.fixture
    def examples_dir(self):
        return os.path.join(os.path.dirname(os.path.dirname(__file__)), 'examples')

    def test_fibonacci_transpiles(self, examples_dir):
        """Test fibonacci.sui transpiles"""
        with open(os.path.join(examples_dir, 'fibonacci.sui')) as f:
            sui_code = f.read()
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(sui_code)
        assert "(module" in code
        assert "(func $f0" in code  # Fibonacci function

    def test_fizzbuzz_transpiles(self, examples_dir):
        """Test fizzbuzz.sui transpiles"""
        with open(os.path.join(examples_dir, 'fizzbuzz.sui')) as f:
            sui_code = f.read()
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(sui_code)
        assert "(module" in code
        assert "(loop $loop" in code  # Should have loop structure

    def test_list_sum_transpiles(self, examples_dir):
        """Test list_sum.sui transpiles with memory"""
        with open(os.path.join(examples_dir, 'list_sum.sui')) as f:
            sui_code = f.read()
        transpiler = Sui2WatTranspiler()
        code = transpiler.transpile(sui_code)
        assert "(module" in code
        assert "(memory 1)" in code  # Should have memory for arrays
        assert "(i32.store)" in code  # Array writes
        assert "(i32.load)" in code   # Array reads




"""
Tests for sui2wasm (Sui to WebAssembly binary compiler)
"""

import pytest
import subprocess
import sys
import os
import tempfile


class TestSui2WasmCLI:
    """Test sui2wasm CLI"""

    def test_help(self):
        """Test --help flag"""
        result = subprocess.run(
            [sys.executable, 'sui2wasm.py', '--help'],
            capture_output=True,
            text=True,
            cwd=os.path.dirname(os.path.dirname(__file__))
        )
        assert result.returncode == 0
        assert 'Usage' in result.stdout

    def test_compile_fibonacci(self):
        """Test compiling fibonacci.sui to wasm"""
        with tempfile.NamedTemporaryFile(suffix='.wasm', delete=False) as f:
            output_path = f.name
        
        try:
            result = subprocess.run(
                [sys.executable, 'sui2wasm.py', 'examples/fibonacci.sui', '-o', output_path],
                capture_output=True,
                text=True,
                cwd=os.path.dirname(os.path.dirname(__file__))
            )
            
            # Check if wat2wasm is available
            if 'wat2wasm not found' in result.stderr:
                pytest.skip("wat2wasm not installed")
            
            assert result.returncode == 0
            assert os.path.exists(output_path)
            
            # Check it's a valid wasm file (magic number)
            with open(output_path, 'rb') as f:
                magic = f.read(4)
                assert magic == b'\x00asm', "Invalid Wasm magic number"
        finally:
            if os.path.exists(output_path):
                os.unlink(output_path)

    def test_compile_fizzbuzz(self):
        """Test compiling fizzbuzz.sui to wasm"""
        with tempfile.NamedTemporaryFile(suffix='.wasm', delete=False) as f:
            output_path = f.name
        
        try:
            result = subprocess.run(
                [sys.executable, 'sui2wasm.py', 'examples/fizzbuzz.sui', '-o', output_path],
                capture_output=True,
                text=True,
                cwd=os.path.dirname(os.path.dirname(__file__))
            )
            
            if 'wat2wasm not found' in result.stderr:
                pytest.skip("wat2wasm not installed")
            
            assert result.returncode == 0
            assert os.path.exists(output_path)
            
            # Verify file size is reasonable
            size = os.path.getsize(output_path)
            assert size > 50, "Wasm file too small"
            assert size < 10000, "Wasm file too large"
        finally:
            if os.path.exists(output_path):
                os.unlink(output_path)

    def test_file_not_found(self):
        """Test error handling for missing file"""
        result = subprocess.run(
            [sys.executable, 'sui2wasm.py', 'nonexistent.sui'],
            capture_output=True,
            text=True,
            cwd=os.path.dirname(os.path.dirname(__file__))
        )
        assert result.returncode == 1
        assert 'not found' in result.stderr.lower()

    def test_default_output_name(self):
        """Test default output filename"""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a test sui file
            sui_path = os.path.join(tmpdir, 'test.sui')
            with open(sui_path, 'w') as f:
                f.write('= g0 42\n. g0')
            
            result = subprocess.run(
                [sys.executable, 'sui2wasm.py', sui_path],
                capture_output=True,
                text=True,
                cwd=os.path.dirname(os.path.dirname(__file__))
            )
            
            if 'wat2wasm not found' in result.stderr:
                pytest.skip("wat2wasm not installed")
            
            # Should create test.wasm in same directory
            expected_output = os.path.join(tmpdir, 'test.wasm')
            assert result.returncode == 0
            assert os.path.exists(expected_output)


class TestSui2WasmModule:
    """Test sui2wasm module functions"""

    def test_compile_to_wasm_simple(self):
        """Test compile_to_wasm with simple code"""
        try:
            from sui2wasm import compile_to_wasm
        except ImportError:
            # Add parent directory to path
            sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
            from sui2wasm import compile_to_wasm
        
        wasm_bytes = compile_to_wasm('= g0 42\n. g0')
        
        if wasm_bytes is None:
            pytest.skip("wat2wasm not installed")
        
        assert isinstance(wasm_bytes, bytes)
        assert wasm_bytes[:4] == b'\x00asm'  # Wasm magic number

    def test_compile_to_wasm_with_function(self):
        """Test compile_to_wasm with function definition"""
        try:
            from sui2wasm import compile_to_wasm
        except ImportError:
            sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))
            from sui2wasm import compile_to_wasm
        
        code = """
# 0 1 {
+ v0 a0 1
^ v0
}
= g0 10
$ g1 0 g0
. g1
"""
        wasm_bytes = compile_to_wasm(code)
        
        if wasm_bytes is None:
            pytest.skip("wat2wasm not installed")
        
        assert isinstance(wasm_bytes, bytes)
        assert len(wasm_bytes) > 50




#!/usr/bin/env python3
"""
Sui to WebAssembly Compiler

Compiles Sui code to WebAssembly Text Format (WAT) and binary (.wasm).
"""

import sys
import subprocess
import tempfile
import os
from typing import Optional

def get_version() -> str:
    """Get package version"""
    try:
        from importlib.metadata import version
        return version("sui-lang")
    except Exception:
        return "0.4.1"


class Sui2WatTranspiler:
    """Sui to WAT transpiler"""

    def __init__(self):
        self.output: list[str] = []
        self.indent = 0
        self.functions: dict[int, dict] = {}
        self.used_globals: set[int] = set()
        self.use_memory = False

    def emit(self, line: str):
        """Emit a line with proper indentation"""
        self.output.append("  " * self.indent + line)

    def parse_line(self, line: str) -> list[str] | None:
        """Parse a single line"""
        if ';' in line:
            line = line[:line.index(';')]
        line = line.strip()
        if not line:
            return None
        
        tokens = []
        i = 0
        while i < len(line):
            if line[i] in ' \t':
                i += 1
                continue
            if line[i] == '"':
                j = i + 1
                while j < len(line) and line[j] != '"':
                    if line[j] == '\\':
                        j += 2
                    else:
                        j += 1
                tokens.append(line[i:j + 1])
                i = j + 1
                continue
            j = i
            while j < len(line) and line[j] not in ' \t':
                j += 1
            tokens.append(line[i:j])
            i = j
        return tokens if tokens else None

    def collect_info(self, lines: list[list[str]]):
        """Collect used variables and check if memory is needed"""
        for tokens in lines:
            if not tokens:
                continue
            op = tokens[0]
            if op in ['[', ']', '{']:
                self.use_memory = True
            for token in tokens[1:]:
                if token.startswith('g'):
                    try:
                        self.used_globals.add(int(token[1:]))
                    except ValueError:
                        pass

    def resolve_value(self, val: str) -> str:
        """Resolve a value to WAT code"""
        if val.startswith('v'):
            idx = int(val[1:])
            return f"(local.get $v{idx})"
        elif val.startswith('g'):
            idx = int(val[1:])
            return f"(global.get $g{idx})"
        elif val.startswith('a'):
            idx = int(val[1:])
            return f"(local.get $a{idx})"
        elif val.startswith('"'):
            return "(i32.const 0)"
        elif '.' in val:
            return f"(i32.const {int(float(val))})"
        else:
            try:
                return f"(i32.const {int(val)})"
            except ValueError:
                return "(i32.const 0)"

    def set_var(self, var: str) -> str:
        """Generate WAT instruction to set a variable"""
        if var.startswith('v'):
            idx = int(var[1:])
            return f"(local.set $v{idx})"
        elif var.startswith('g'):
            idx = int(var[1:])
            return f"(global.set $g{idx})"
        return ""

    def transpile_block(self, lines: list[list[str]], local_vars: set[int], is_function: bool = False) -> list[str]:
        """Transpile a block using state machine pattern for labels/jumps"""
        result = []
        
        labels: set[int] = set()
        for tokens in lines:
            if tokens and tokens[0] == ':':
                labels.add(int(tokens[1]))

        if labels:
            state_map: dict[int, int] = {}
            state_num = 1
            for label in sorted(labels):
                state_map[label] = state_num
                state_num += 1

            states: dict[int, list] = {0: []}
            current_state = 0
            
            for tokens in lines:
                if not tokens:
                    continue
                if tokens[0] == ':':
                    label_id = int(tokens[1])
                    current_state = state_map[label_id]
                    if current_state not in states:
                        states[current_state] = []
                elif tokens[0] == '}':
                    continue
                else:
                    if current_state not in states:
                        states[current_state] = []
                    states[current_state].append(tokens)

            result.append("(local $_state i32)")
            result.append("(local.set $_state (i32.const 0))")
            result.append("(block $exit")
            result.append("  (loop $loop")
            
            for state_id in sorted(states.keys()):
                result.append(f"    ;; State {state_id}")
                result.append(f"    (if (i32.eq (local.get $_state) (i32.const {state_id}))")
                result.append("      (then")
                
                state_lines = states[state_id]
                has_jump = False
                
                for tokens in state_lines:
                    if tokens[0] == '?':
                        cond_code = self.resolve_value(tokens[1])
                        target_label = int(tokens[2])
                        target_state = state_map.get(target_label, 0)
                        result.append(f"        {cond_code}")
                        result.append(f"        (if")
                        result.append(f"          (then")
                        result.append(f"            (local.set $_state (i32.const {target_state}))")
                        result.append(f"            (br $loop)")
                        result.append(f"          )")
                        result.append(f"        )")
                    elif tokens[0] == '@':
                        target_label = int(tokens[1])
                        target_state = state_map.get(target_label, 0)
                        result.append(f"        (local.set $_state (i32.const {target_state}))")
                        result.append(f"        (br $loop)")
                        has_jump = True
                    elif tokens[0] == '^':
                        val_code = self.resolve_value(tokens[1])
                        result.append(f"        {val_code}")
                        result.append(f"        (return)")
                        has_jump = True
                    elif tokens[0] == '$':
                        result_var = tokens[1]
                        func_id = tokens[2]
                        args = tokens[3:]
                        for arg in args:
                            result.append(f"        {self.resolve_value(arg)}")
                        result.append(f"        (call $f{func_id})")
                        result.append(f"        {self.set_var(result_var)}")
                    else:
                        insts = self.transpile_instruction(tokens)
                        for inst in insts:
                            result.append(f"        {inst}")
                
                if not has_jump:
                    next_state = state_id + 1
                    if next_state in states:
                        result.append(f"        (local.set $_state (i32.const {next_state}))")
                        result.append(f"        (br $loop)")
                    else:
                        result.append(f"        (br $exit)")
                
                result.append("      )")
                result.append("    )")
            
            result.append("    (br $exit)")
            result.append("  )")
            result.append(")")
        else:
            for tokens in lines:
                if not tokens or tokens[0] == '}':
                    continue
                if tokens[0] == '^':
                    val_code = self.resolve_value(tokens[1])
                    result.append(val_code)
                    result.append("(return)")
                elif tokens[0] == '$':
                    result_var = tokens[1]
                    func_id = tokens[2]
                    args = tokens[3:]
                    for arg in args:
                        result.append(self.resolve_value(arg))
                    result.append(f"(call $f{func_id})")
                    result.append(self.set_var(result_var))
                else:
                    insts = self.transpile_instruction(tokens)
                    for inst in insts:
                        result.append(inst)

        return result

    def transpile_instruction(self, tokens: list[str]) -> list[str]:
        """Transpile a single instruction to WAT"""
        if not tokens:
            return []
        
        op = tokens[0]
        result = []

        if op == '=':
            val_code = self.resolve_value(tokens[2])
            result.append(val_code)
            result.append(self.set_var(tokens[1]))

        elif op in ['+', '-', '*', '/', '%']:
            a_code = self.resolve_value(tokens[2])
            b_code = self.resolve_value(tokens[3])
            op_map = {'+': 'i32.add', '-': 'i32.sub', '*': 'i32.mul', '/': 'i32.div_s', '%': 'i32.rem_s'}
            result.append(a_code)
            result.append(b_code)
            result.append(f"({op_map[op]})")
            result.append(self.set_var(tokens[1]))

        elif op == '<':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.lt_s)")
            result.append(self.set_var(tokens[1]))

        elif op == '>':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.gt_s)")
            result.append(self.set_var(tokens[1]))

        elif op == '~':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.eq)")
            result.append(self.set_var(tokens[1]))

        elif op == '!':
            result.append(self.resolve_value(tokens[2]))
            result.append("(i32.eqz)")
            result.append(self.set_var(tokens[1]))

        elif op == '&':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.and)")
            result.append(self.set_var(tokens[1]))

        elif op == '|':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.or)")
            result.append(self.set_var(tokens[1]))

        elif op == '.':
            result.append(self.resolve_value(tokens[1]))
            result.append("(call $print_i32)")

        elif op == '[':
            size_code = self.resolve_value(tokens[2])
            result.append("(global.get $heap_ptr)")
            result.append(self.set_var(tokens[1]))
            result.append("(global.get $heap_ptr)")
            result.append(size_code)
            result.append("(i32.const 4)")
            result.append("(i32.mul)")
            result.append("(i32.add)")
            result.append("(global.set $heap_ptr)")
            self.use_memory = True

        elif op == ']':
            result.append(self.resolve_value(tokens[2]))
            result.append(self.resolve_value(tokens[3]))
            result.append("(i32.const 4)")
            result.append("(i32.mul)")
            result.append("(i32.add)")
            result.append("(i32.load)")
            result.append(self.set_var(tokens[1]))
            self.use_memory = True

        elif op == '{':
            if len(tokens) >= 4:
                result.append(self.resolve_value(tokens[1]))
                result.append(self.resolve_value(tokens[2]))
                result.append("(i32.const 4)")
                result.append("(i32.mul)")
                result.append("(i32.add)")
                result.append(self.resolve_value(tokens[3]))
                result.append("(i32.store)")
                self.use_memory = True

        return result

    def transpile_function(self, func_id: int, argc: int, body: list[list[str]]) -> list[str]:
        """Transpile a function to WAT"""
        result = []
        
        local_vars: set[int] = set()
        for tokens in body:
            if not tokens:
                continue
            for token in tokens:
                if token.startswith('v'):
                    try:
                        local_vars.add(int(token[1:]))
                    except ValueError:
                        pass

        params = " ".join(f"(param $a{i} i32)" for i in range(argc))
        result.append(f'(func $f{func_id} (export "f{func_id}") {params} (result i32)')
        
        for v in sorted(local_vars):
            result.append(f"  (local $v{v} i32)")
        
        body_code = self.transpile_block(body, local_vars, is_function=True)
        for line in body_code:
            result.append(f"  {line}")
        
        result.append("  (i32.const 0)")
        result.append(")")
        
        return result

    def transpile(self, code: str) -> str:
        """Transpile Sui code to WAT"""
        self.output = []
        self.used_globals = set()
        self.use_memory = False
        self.functions = {}
        
        lines_raw = code.strip().split('\n')
        lines = []
        for line in lines_raw:
            parsed = self.parse_line(line)
            if parsed:
                lines.append(parsed)

        self.collect_info(lines)

        i = 0
        while i < len(lines):
            if lines[i][0] == '#':
                func_id = int(lines[i][1])
                argc = int(lines[i][2])
                body = []
                i += 1
                depth = 1
                while i < len(lines) and depth > 0:
                    if lines[i][0] == '#':
                        depth += 1
                    elif lines[i][0] == '}':
                        depth -= 1
                        if depth == 0:
                            break
                    body.append(lines[i])
                    i += 1
                self.functions[func_id] = {'argc': argc, 'body': body}
                self.collect_info(body)
            i += 1

        self.emit("(module")
        self.indent += 1

        self.emit(";; External function imports")
        self.emit('(import "env" "print_i32" (func $print_i32 (param i32)))')
        self.emit("")

        if self.use_memory:
            self.emit(";; Linear memory for arrays")
            self.emit("(memory 1)")
            self.emit("(global $heap_ptr (mut i32) (i32.const 0))")
            self.emit("")

        if self.used_globals:
            self.emit(";; Global variables")
            for g in sorted(self.used_globals):
                self.emit(f"(global $g{g} (export \"g{g}\") (mut i32) (i32.const 0))")
            self.emit("")

        if self.functions:
            self.emit(";; Function definitions")
            for func_id, func_info in sorted(self.functions.items()):
                func_lines = self.transpile_function(func_id, func_info['argc'], func_info['body'])
                for line in func_lines:
                    self.emit(line)
                self.emit("")

        self.emit(";; Main function")
        self.emit("(func $main (export \"main\") (result i32)")
        self.indent += 1

        main_locals: set[int] = set()
        main_lines = []
        i = 0
        while i < len(lines):
            if lines[i][0] == '#':
                depth = 1
                i += 1
                while i < len(lines) and depth > 0:
                    if lines[i][0] == '#':
                        depth += 1
                    elif lines[i][0] == '}':
                        depth -= 1
                    i += 1
            else:
                main_lines.append(lines[i])
                for token in lines[i]:
                    if token.startswith('v'):
                        try:
                            main_locals.add(int(token[1:]))
                        except ValueError:
                            pass
                i += 1

        for v in sorted(main_locals):
            self.emit(f"(local $v{v} i32)")

        body_code = self.transpile_block(main_lines, main_locals)
        for line in body_code:
            self.emit(line)

        self.emit("(i32.const 0)")
        self.indent -= 1
        self.emit(")")

        self.indent -= 1
        self.emit(")")

        return '\n'.join(self.output)


def compile_to_wasm(sui_code: str) -> bytes | None:
    """Compile Sui code to Wasm binary"""
    transpiler = Sui2WatTranspiler()
    wat_code = transpiler.transpile(sui_code)
    
    try:
        with tempfile.NamedTemporaryFile(mode='w', suffix='.wat', delete=False) as f:
            f.write(wat_code)
            wat_path = f.name
        
        wasm_path = wat_path.replace('.wat', '.wasm')
        
        result = subprocess.run(
            ['wat2wasm', wat_path, '-o', wasm_path],
            capture_output=True,
            text=True
        )
        
        if result.returncode != 0:
            print(f"Error: wat2wasm failed: {result.stderr}", file=sys.stderr)
            return None
        
        with open(wasm_path, 'rb') as f:
            wasm_bytes = f.read()
        
        os.unlink(wat_path)
        os.unlink(wasm_path)
        
        return wasm_bytes
        
    except FileNotFoundError:
        print("Error: wat2wasm not found. Install wabt: brew install wabt", file=sys.stderr)
        return None
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return None


def main():
    if len(sys.argv) >= 2 and sys.argv[1] in ('--version', '-V'):
        print(f"sui-lang {get_version()}")
        return
    
    if len(sys.argv) < 2 or sys.argv[1] in ['-h', '--help']:
        print("Usage: sui2wasm <input.sui> [-o <output.wasm>]")
        print()
        print("Compile Sui code to WebAssembly binary.")
        print()
        print("Options:")
        print("  -o FILE      Output file (default: input.wasm)")
        print("  --version    Show version")
        print()
        print("Requirements:")
        print("  wat2wasm     Install via: brew install wabt")
        sys.exit(0 if '-h' in sys.argv or '--help' in sys.argv else 1)
    
    input_file = sys.argv[1]
    
    output_file = input_file.rsplit('.', 1)[0] + '.wasm'
    if '-o' in sys.argv:
        idx = sys.argv.index('-o')
        if idx + 1 < len(sys.argv):
            output_file = sys.argv[idx + 1]
    
    try:
        with open(input_file, 'r') as f:
            sui_code = f.read()
    except FileNotFoundError:
        print(f"Error: File not found: {input_file}", file=sys.stderr)
        sys.exit(1)
    
    wasm_bytes = compile_to_wasm(sui_code)
    
    if wasm_bytes is None:
        sys.exit(1)
    
    with open(output_file, 'wb') as f:
        f.write(wasm_bytes)
    
    print(f"✓ Compiled to {output_file} ({len(wasm_bytes)} bytes)")


if __name__ == '__main__':
    main()
